//
//  ViewController.m
//  PrintTest
//
//  Created by michael hall on 7/24/15.
//  Copyright (c) 2015 Boca Systems. All rights reserved.
//
//  This application is intended to demonstrate printing files through the Boca Systems printer driver
//  and allowing the port monitor to read returning status from the Boca Systems USB printer.
//
//  This application will focus on the Default Printer, so make sure a Boca Systems Ticket is selected as the default
//
//  Below you will see examples of how to transmit plain text, BMP, JPEG, PDF, RTF and raw text files to a Boca Systems
//  printer through a Boca Systems driver and read the returning status.
//
//  Status will be returned with each job submitted, not necessarily with each ticket printed.  In most examples below
//  one ticket is printed with each job and a ticket ACK can be read afterwards.  In the PDF example below six tickets are printed in one job.
//  The driver task/job will complete before all the tickets are printed so the monitor won't necessarily pick up all six Ticket ACKs
//  since the printer does not generate the Ticket ACK until after the ticket has physically printed.

#import "ViewController.h"
#import <ApplicationServices/ApplicationServices.h>
#include <cups/cups.h>
#include <cups/ppd.h>
#include <cups/raster.h>
#include <cups/sidechannel.h>
#include <cups/backend.h>
#include <stdio.h>
#include <dlfcn.h>

@implementation ViewController

void* raster_handle = NULL;
typedef int (*receive_data_t)(char *buffer, char *SerialNumber);
typedef int (*start_read_t)(char *SerialNumber);
typedef int (*stop_read_t)(char *SerialNumber);

@synthesize textOut, scrollOut, FileType;
@synthesize PrinterName, PrinterID, PrinterSN;

- (void)viewDidLoad
{
    [super viewDidLoad];
    // Do any additional setup after loading the view.
    [self FindDefaultSN];               //Establish which printer is the default printer
    [self ClearScreen];                 //setup screen to display infomation and clear screen of any previous status
    
}

- (void)viewDidUnload
{
    // Release any retained subviews of the main view.
}


- (void)setRepresentedObject:(id)representedObject
{
    [super setRepresentedObject:representedObject];
    // Update the view, if already loaded.
}

- (NSPrintInfo *)printInfo
{
    if (_printInfo == nil)
    {
        _printInfo = [[NSPrintInfo alloc] init];
    }
    return _printInfo;
}

- (IBAction)SelectFile:(id)sender
{    
    CFURLRef fileURL = NULL;
    CFStringRef mimeType = NULL;
    NSBundle *mainBundle = [NSBundle mainBundle];
    Boolean done=true;
    
    // Obtain the print settings from the printInfo.
    PMPrintSettings printSettings = [self.printInfo PMPrintSettings];
    //clear job name from previous jobs.  This is important if using the EOT job name signal demonstrated in case 6 below
    PMPrintSettingsSetJobName(printSettings,(CFStringRef)@"");
    do
    {
        switch ([[FileType selectedCell] tag])
        {
        
        case 1:
            //NSLog(@"Plain Text");
            //This is used as a standard text file and will print the text conatined in the file unto the ticket
            PMPrintSettingsSetJobName(printSettings,(CFStringRef)@"Plain Text Test");
            fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"plain" withExtension:@"txt"];
            mimeType = CFSTR("text/plain");                         //Identify file type as a plain text document
            break;
            
        case 2:
            //NSLog(@"BMP");
            //The Boca Systems printer only prints Black, so any BMP image file should
            //be saved in an editor as a monochrome B&W file before sending to the printer.  Color BMPs can
            //be printed through the driver but they will be rasterized which causes some degregation of the image.
            
            //For image files, tell the printing system to fit the image to the page.
            PMPrintSettingsSetValue(printSettings, kPMFitToPageKey, CFSTR("true"), false);
            PMPrintSettingsSetJobName(printSettings,(CFStringRef)@"BMP Image Test");

            //Identify file in fileURL to be printed.  3 BMP choices shown below.
            //fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"testImage" withExtension:@"bmp"];               //Color image, which gets rasterized by driver to B&W
            //fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"lemur_bw" withExtension:@"bmp"];                //Monochrome B&W image
            fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"checkmark" withExtension:@"bmp"];                 //Monochrome B&W image
            mimeType = CFSTR("image/x-bmp");                                                                            //Identify file type as BMP
            break;
            
        case 3:
            //NSLog(@"JPEG");
            //The Boca Systems printer only prints Black, so any JPEG image file should
            //be saved in an editor as a monochrome B&W file before sending to the printer.  Color JPEGs can
            //be printed through the driver but they will be rasterized which causes some degregation of the image.
            
            //For image files, tell the printing system to fit the image to the page.
            PMPrintSettingsSetValue(printSettings, kPMFitToPageKey, CFSTR("true"), false);
            PMPrintSettingsSetJobName(printSettings,(CFStringRef)@"JPEG Image Test");
            
            //Identify file in fileURL to be printed.  3 JPEG choices shown below.
            //fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"testImage" withExtension:@"jpg"];               //Color image, which gets rasterized by driver to B&W
            //fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"lemur_bw" withExtension:@"jpg"];                //Monochrome B&W image
            fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"checkmark" withExtension:@"jpg"];                 //Monochrome B&W image
            mimeType = CFSTR("image/jpeg");                                                                             //Identify file type as JPEG
            break;
            
        case 4:
            //NSLog(@"PDF");
            PMPrintSettingsSetValue(printSettings, kPMFitToPageKey, CFSTR("true"), false);
            PMPrintSettingsSetJobName(printSettings,(CFStringRef)@"PDF File Test");
            fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"concert_stock_landscape_6pack" withExtension:@"pdf"];
            mimeType = CFSTR("application/pdf");                                                                        //Identify file type as a PDF document
            break;
            
        case 5:
            //NSLog(@"RTF");
            PMPrintSettingsSetValue(printSettings, kPMFitToPageKey, CFSTR("true"), false);
            PMPrintSettingsSetJobName(printSettings,(CFStringRef)@"RTF File Test");
            fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"BocaTest" withExtension:@"rtf"];
            mimeType = CFSTR("text/rtf");                                                                               //Identify file type as a rich text document
            break;
            
        case 6:
            //NSLog(@"Raw Text File");
            //This is used when the raw FGL text contained in the file is to be interpretted by the printer, not printed.
            //The text is typically FGL commands, mixed with plain text as demonstrated in the three files listed below.
            //At release time the config_fgl file is being used
            //
            //When transmitting a vnd.cups-raw mime file type the raw data is sent straight through the spooler to the printer
            //without the filter/driver being used.  Without the driver being used, the printer monitor does not get executed to read
            //data returning from the printer.  That is fine if the application does not need to report the return data.
            //
            //Should however the application needs the return data, then a second transmission is needed which uses the driver/monitor.
            //Below there is a second file transmitted to the printer to achieve that.  This file is in a plain text format.  There
            //is one byte of data in the eot.txt file used.  The byte is a hex 3 which represents End of transmission.  This will
            //not cause the boca systems printer to do anything, while allowing the printer monitor to run and check for data returning
            //from the printer due to the previous raw text file sent.  The key point here is to set the Job Name to EOT as shown
            //below.  The printer driver will look for this job name and use this as a signal and not generate a form
            //feed which stops the printer from ejecting a blank ticket needlessly.
            //
            //The "done" flag is a boolean used to get both files sent in this one case
            if(done)
            {
                //Identify file in fileURL to be printed.  3 raw text choices shown below.
                PMPrintSettingsSetJobName(printSettings,(CFStringRef)@"FGL Raw Text Test");
                fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"config_fgl" withExtension:@"txt"];    //FGL most popular configuration items on one ticket
                //fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"qrtest" withExtension:@"txt"];      //FGL QR barcode text commands This file will print 12 tickets
                //fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"fgl_test" withExtension:@"txt"];    //FGL Standard Test ticket
                mimeType = CFSTR("application/vnd.cups-raw");                                                   //Identify file type as raw text
                done = false;
            }
            else
            {
                PMPrintSettingsSetJobName(printSettings,(CFStringRef)@"EOT");                                   //signal driver for no form feed needed
                fileURL = (__bridge CFURLRef)[mainBundle URLForResource:@"eot" withExtension:@"txt"];
                mimeType = CFSTR("text/plain");                                                                 //Identify file type as a plain text document
                done = true;
            }
            
            
            break;
            
        }
    
        // Obtain the print session from the printInfo.
        PMPrintSession printSession = [self.printInfo PMPrintSession];
        PMDestinationType printDestination = 0;
        OSStatus status = noErr;
    
        // Verify that the destination is the printer.
        status = PMSessionGetDestinationType(printSession, printSettings, &printDestination);
        if ((status != noErr) || (printDestination != kPMDestinationPrinter))
        {
        
            NSLog(@"Either got an error from PMSessionGetDestinationType or the print destination wasn't kPMDestinationPrinter");
            return;
        }
    
        // The destination printer is needed by PMPrinterPrintWithFile and related functions.
        PMPrinter currentPrinter = NULL;
        status = PMSessionGetCurrentPrinter(printSession, &currentPrinter);
        if (status != noErr)
        {
        
            NSLog(@"Got an error from PMSessionGetCurrentPrinter (%d)", status);
            return;
        }
    
        //One reason PMPrinterPrintWithFile may fail is if the specified printer can not handle the file's mime type. Use PMPrinterGetMimeTypes() to check whether a mime type is supported.
        CFArrayRef mimeTypes;
        status = PMPrinterGetMimeTypes(currentPrinter, printSettings, &mimeTypes);
        if (status == noErr && mimeTypes != NULL)
        {
        
            CFIndex arrayCount = CFArrayGetCount(mimeTypes);
            if (CFArrayContainsValue(mimeTypes, CFRangeMake(0, arrayCount), mimeType))
            {
            
                // Obtain the page format from the printInfo.
                PMPageFormat pageFormat =  [self.printInfo PMPageFormat];
            
                status = PMPrinterPrintWithFile(currentPrinter, printSettings, pageFormat, mimeType, fileURL);
                if (status != noErr)
                {
                    NSLog(@"PMPrinterPrintWithFile returned error %d", status);
                }
            }
            else
            {
            
                NSLog(@"MIME type %@ isn't supported by the printer", (__bridge NSString *)mimeType);
            }
        }
        
    }while(!done);
    
}

//This routine will call Receive Data method contained in the driver to return all data
//that has been monitored returning from the printer.  The port monitor holds all the data
//until the call below is made.  A user can call this in between each ticket printed or
//once in a while depending upon their need
- (IBAction)ReadStatus:(id)sender
{
    char data[1024];
    int bytecount=0;

    NSString *s;
    NSString* statusString=[[NSMutableString alloc] init];
    int i=0;
    char x;
    
    /*
     RTLD_LOCAL
     RTLD_LAZY
     RTLD_GLOBAL
     RTLD_DEFAULT
     RTLD_FIRST
     RTLD_NOW
     */
    
    //Use this to create a link to the Boca Systems driver filter.  This is used when calling
    //the receive_data method to retrieve returning data from the printer.
    raster_handle = dlopen("/Library/Printers/Boca/Filters/rastertobocagray", RTLD_NOW);
    if(raster_handle)
    {
        receive_data_t recevice_data = (receive_data_t)dlsym(raster_handle, "ReceiveData");
        if(recevice_data)
        {
            dlerror();
            //Pass two parameters when calling receive data
            // 1. pointer to buffer, for the data to be written
            // 2. Serial number of Boca Systems printer to poll
            bytecount = (int)recevice_data(data,printerSN);
        }
        //@msh temp test dlclose(raster_handle);
    }
    
    //Use this code to decipher data returning.  Data can be a one byte status ASCII value which is less than 32.  Or the data
    // can be a printable string ASCII value which is greater than 31.
    if( bytecount > 0 )
    {
        statusString = @"";                                                 //blank out status string
        for (i=0; i < bytecount; i++)
        {
            x = data[i];
            //check if character found is printable string or control status
            if(x < 32)
            {
                //check if status string contains something to output
                if(![statusString isEqual: @""])
                {
                    //output string to display
                    [self DisplayText:statusString];
                    statusString = @"";                                     //blank out status string
                }
                
                //now process control character to establish status meaning of most common status
                switch (x)
                {
                    case 0x6:
                        s = @"Ticket ACK";
                        break;
                    
                    case 0x8:
                        s = @"Invalid Checksum";
                        break;
                        
                    case 0x9:
                        s = @"Valid Checksum";
                        break;
                        
                    case 0x10:
                        s = @"Out of Tickets";
                        break;
                    
                    case 0x11:
                        s = @"X-On";
                        break;
                    
                    case 0x12:
                        s = @"Power On";
                        break;
                    
                    case 0x13:
                        s = @"X-Off";
                        break;
                    
                    case 0x15:
                        s = @"Ticket NAK";
                        break;
                    
                    case 0x18:
                        s = @"Ticket Jam";
                        break;
                    
                    case 0x19:
                        s = @"Illegal Data";
                        break;
                    
                    case 0x1D:
                        s = @"Cutter Jam";
                        break;
                    
                    default:
                        s = @"";                //Undetermined Status Returned
                        break;
                }
                
                if(![s isEqualToString:@""])
                    [self DisplayText:s];
            }
            else //must be ascii character so add to status string
            {
                statusString = [NSString stringWithFormat:@"%@%c",statusString,x];
            }
            
        }
        //check if status string contains something to output
        if(![statusString isEqual: @""])
        {
            //output string to display
            [self DisplayText:statusString];
            statusString = @"";                                     //blank out status string
        }
        
    }
    else
    {
        if (bytecount == -1)
            [self DisplayText:@"No printer Data Found"];
    }

//textOut.selectedRange = NSMakeRange(textOut.text.length - 1, 0);    //implement auto-scroll to bottom of screen

}

- (IBAction)RefreshScreen:(id)sender
{

    _printInfo = nil;                   //Clear setting for currently selected printer, and then reestablish in printInfo
    [self FindDefaultSN];               //Establish which printer is the default printer
    [self ClearScreen];                 //setup screen to display infomation and clear screen of any previous status

}

- (void)DisplayText:(NSString *)data
{
    //define string and initialize it with what is already on the screen
    NSString *text=StatusString;
    
    //concatenate what was on the display screen with the new string just passed in and add a line feed
    StatusString = [text stringByAppendingFormat:@"%@%@",data,@"\n"];

    //display strings in scrollable area
    [textOut setString:StatusString];
    
}

-(void) ClearScreen
{
    //display strings
    [PrinterName setStringValue:(NSString *)CFBridgingRelease(printerName)];
    [PrinterID setStringValue:(NSString *)CFBridgingRelease(printerID)];

    //define string and initialize it with what is already on the screen
    NSString *text=@"";
    NSString *SN=@"";
    SN = [text stringByAppendingFormat:@"%s",printerSN];
    
    //display strings in scrollable area
    [PrinterSN setStringValue:SN];
    
    //clear text in scrollable area
    [textOut setString:@""];

    StatusString = @"";
}

//This application will be used to find the serial number associated with the printer currently being used.
//The serial number of the printer is used when reading data back from the printer
-(void) FindDefaultSN
{
    //Make a list from the available printers
    const char *sn;
    char *pos;
    
    CFArrayRef printerList;
    // Don't forget to check for errors in production code!
    PMServerCreatePrinterList(kPMServerLocal, &printerList);
    

    long numberOfPrinters = CFArrayGetCount(printerList);
    long printerIndex;
    
    //Check each printer
    for(printerIndex = 0; printerIndex < numberOfPrinters; printerIndex++)
    {
        
        //Make a reference to the printer
        PMPrinter printer =(PMPrinter)CFArrayGetValueAtIndex(printerList, printerIndex);
        
        //Check to see if this is the default printer. If not skip it
        Boolean printerIsDefault = PMPrinterIsDefault(printer);
        if(printerIsDefault)
        {
            //Find out its name
            printerName = PMPrinterGetName(printer);
            
            //Note the Printer Name is different from the queue name.
            printerID = PMPrinterGetID(printer);
            
            // Get the URI (Uniform Resource Identifier).
            CFURLRef printerURI;
            PMPrinterCopyDeviceURI(printer, &printerURI);
            
            //Read the URI string from the system information and build a string from that.
            //The Boca Systems printer responds with a string read similar to --> "usb://BOCA%20SYSTEMS/46%20300?serial=0000123456"
            sn = [(NSString *)CFURLGetString(printerURI) UTF8String];

            //Find serial number string of Boca Systems printer contained within string
            pos = strstr(sn,"serial=0000");
            if(pos)
                strncpy(printerSN,pos+11,6);
            else
                strcpy(printerSN,"-1");
            
            printerIndex = numberOfPrinters + 1;
            CFRelease(printerURI);

        }
    }
}

@end

